package haskell.data;

import java.util.*;
import java.util.functions.*;

import haskell.data.functions.*;

import java.io.*;
/**

  This class gives static methods for (almost) all standard List functions
  of the Haskell Prelude in terms of Java 1.8.

  @author Sven Eric Panitz
  @mail sveneric@panitz.name
  @institution Hochschule RheinMain
 **/

public class Iterables {
    /**
       This class is the key for creating an Iterable for a given Iterator. The class is used
       by the method »make«. 
       It solves the problem
       of the quick and dirty method »drain«, which runs the complete iterator in order to 
       capture its elements in a List. The method »drain« does not work for infinite Iterator
       ranges.
       This class will work a bit like a CAF list in Haskell. It memorizes allready iterated elements
       in a List. The  
     **/
  static private class CacheIterator<T> implements Iterator<T>{
    Iterator<T> originalIterator;    
    List<T> xs = new ArrayList<>();
    int i = 0;

    public CacheIterator(Iterator<T> originalIterator){
      this.originalIterator=originalIterator;
    }

    public CacheIterator(Iterator<T> originalIterator,List<T> xs){
      this.originalIterator=originalIterator;
      this.xs=xs;
      i=0;
      //      new Exception().printStackTrace();
    }

    public CacheIterator<T> resetedIterator(){
      return new CacheIterator<T>(originalIterator,xs);
    }

    public boolean hasNext(){
      return (i<xs.size() || originalIterator.hasNext());
    }     
    public T next(){
      if (i<xs.size()){
        T result=xs.get(i);
        i++;
        return result;
      }else{
        T result = originalIterator.next();
        xs.add(result);
        i++;
        return result;
      }
    }
  }

  /** 
  This method creates an Iterable for a given Iterator. 
  **/
  private static <T> Iterable<T> make(final Iterator<T> iterator){
    return 
      new Iterable<T>(){
        CacheIterator<T> cached = null;
        @Override public Iterator<T> iterator(){
          if (cached==null) cached=new CacheIterator<>(iterator);
          return cached.resetedIterator();            
        }

	public String toString(){
          if (isEmpty()){
            return "[]";
          }else{
            StringBuffer result= new StringBuffer("");  
            T hd = head(this);

            //Iterable<Character> is printed like a String constant.
            final boolean isString = hd instanceof Character; 
            result.append(isString?"\"":"["); 
	    result.append(hd);
            tail(this).forEach((x)->{result.append((isString?"":",")+x);});

            result.append(isString?"\"":"]");
            return result.toString();
          }
        }
      };
  }

  public static <T> Iterable<T> make(final Iterable<T> iterable){
    return make(iterable.iterator());
  }

  public static Iterable<Character> mkStringIter(String str){
    final char[] cs = str.toCharArray();
    Iterator<Character> it = new Iterator<Character>() {
      int i= 0;
      public boolean hasNext(){return i<cs.length;}
      public Character next(){
        if (!hasNext())throw new HeadForEmptyIterableException("TupleTest.mkStringIter: empty list");
        i++;
        return cs[i-1];
      }
    }; 
    return make(it);
  }

  static public class HeadForEmptyIterableException extends RuntimeException{
    public HeadForEmptyIterableException(String explanation){
      super(explanation);
    }
  }

  public static class Nil<T> implements Iterable<T>{
   public String toString(){return"[]";}
   public Iterator<T> iterator(){
      return new Iterator<T>(){
        public boolean hasNext(){return false;}
        public T next(){throw new NoSuchElementException();}
      };
    }
  }

  public static <T> Iterable<T> nil(){return new Nil<T>();}

  static public <T> Iterable<T> cons(final T head,final Iterable<T> tail){
    return append(new Tuple1<T>(head),tail);
  }

    /**
flip             :: (a -> b -> c) -> b -> a -> c
flip f x y       =  f y x
    **/
    static public  <A,B,C> C flip(Function2<A,B,C> f ,B x,A y){
	return f.eval(y,x);
    }


    /**
       Append two lists, i.e.,

       [x1, ..., xm] ++ [y1, ..., yn] == [x1, ..., xm, y1, ..., yn]
       [x1, ..., xm] ++ [y1, ...] == [x1, ..., xm, y1, ...]

       If the first list is not finite, the result is the first list. 
    **/
  static public <T> Iterable<T> append(final Iterable<T> start,final Iterable<T> end){
    Iterator<T> result = new Iterator<T>(){
      Iterator<T> xs = start.iterator();
      Iterator<T> ys = end.iterator();
      public boolean hasNext(){return xs.hasNext() || ys.hasNext();}
      public T next(){return xs.hasNext()?xs.next():ys.next();}
    };
    return make(result);
  }

    /**
       Extract the first element of a list, which must be non-empty. 
    **/
   public static <T> T head(Iterable<T> xs){
    if (xs.isEmpty()) throw new HeadForEmptyIterableException("Prelude.head: empty list");
    return xs.getFirst();
  }
    /**
       Extract the last element of a list, which must be finite and non-empty. 
    **/
  public static <T> T last(Iterable<T> xs){
    if (xs.isEmpty()) throw new HeadForEmptyIterableException("Prelude.last: empty list");
    Iterable<T> ys = tail(xs);
    if (ys.isEmpty()) return head(xs);
    return last(ys);
  }

    /**
       Extract the elements after the head of a list, which must be non-empty. 
    **/
  static public <T> Iterable<T> tail(final Iterable<T> xs){
    if (xs.isEmpty()) throw new HeadForEmptyIterableException("Prelude.tail: empty list");
    Iterator<T> ys = xs.iterator();
    ys.next();
    return make(ys);
  }

    /**
       Return all the elements of a list except the last one. The list must be non-empty. 
     **/
  static public <T> Iterable<T> init(final Iterable<T> xs){
    if (xs.isEmpty()) throw new HeadForEmptyIterableException("Prelude.init: empty list");
    Iterator<T> iterator = new Iterator<T>(){
      Iterator <T> itOrg = xs.iterator();
      T next = itOrg.next();
      public boolean hasNext(){return itOrg.hasNext();}
      public T next(){
        T result = next;
	if (!hasNext()) throw new HeadForEmptyIterableException("Prelude.init: empty list");
	next=itOrg.next();
	return result;
      }
    };
    return make(iterator);
  }

    /**
       O(n). length returns the length of a finite list as an Int. It is an instance of the more general genericLength, the result type of which may be any kind of number. 
    **/
  public static <T> long length(Iterable<T> iterable) {
      return iterable.count();
  }

    /**
       map f xs is the list obtained by applying f to each element of xs, i.e.,

       map f [x1, x2, ..., xn] == [f x1, f x2, ..., f xn]
       map f [x1, x2, ...] == [f x1, f x2, ...]
     **/
  public static <T1,T2> Iterable<T2> map(final Function1<T1,T2> f, Iterable<T1> xs){
    final Iterator<T1> it1 = xs.iterator();
    Iterator<T2> result = new Iterator<T2>(){
      public boolean hasNext(){return it1.hasNext();}
      public T2 next(){
        return f.eval(it1.next());
      }
    };
    return make(result);       
  }    

    /**
       reverse xs returns the elements of xs in reverse order. xs must be finite. 
    **/
  static public <T> Iterable<T> reverse(final Iterable<T> xs){
    if (xs.isEmpty()) return xs;
    return append(reverse(tail(xs)),cons(head(xs),new Nil<T>()));
  }

    /**
-- | The 'intersperse' function takes an element and a list and
-- \`intersperses\' that element between the elements of the list.
-- For example,
--
-- > intersperse ',' "abcde" == "a,b,c,d,e"
intersperse             :: a -> [a] -> [a]
intersperse _   []      = []
intersperse _   [x]     = [x]
intersperse sep (x:xs)  = x : sep : intersperse sep xs
    **/
  static public <T> Iterable<T> intersperse(T sep,final Iterable<T> xs){
    if (xs.isEmpty()) return xs;
    Iterable<T> tl = tail(xs);
    if (tl.isEmpty()) return xs;
    return cons(head(xs),cons(sep,intersperse(sep,tl)));
  }    

    /**
-- | 'intercalate' @xs xss@ is equivalent to @('concat' ('intersperse' xs xss))@.
-- It inserts the list @xs@ in between the lists in @xss@ and concatenates the
-- result.
intercalate :: [a] -> [[a]] -> [a]
intercalate xs xss = concat (intersperse xs xss)
    **/
  static public <T> Iterable<T> intercalate(Iterable<T> xs,final Iterable<Iterable<T>> xss){
      return concat (intersperse (xs, xss));
  }

    /**
-- | The 'transpose' function transposes the rows and columns of its argument.
-- For example,
--
-- > transpose [[1,2,3],[4,5,6]] == [[1,4],[2,5],[3,6]]

transpose               :: [[a]] -> [[a]]
transpose []             = []
transpose ([]   : xss)   = transpose xss
transpose ((x:xs) : xss) = (x : [h | (h:_) <- xss]) : transpose (xs : [ t | (_:t) <- xss])
    **/
  static public <T> Iterable<Iterable<T>> transpose(Iterable<Iterable<T>> xss){
    if (xss.isEmpty()) return xss;    
    Iterable<T> ys=head(xss);
    if (ys.isEmpty()) return transpose(tail(xss));
    T x = head(ys);
    Iterable<T> xs = tail(ys);
    return cons(cons(x,map(Iterables::head,xss)),transpose(cons(xs,map(Iterables::tail,xss))));
  }


    /**
The subsequences function returns the list of all subsequences of the argument.

 subsequences "abc" == ["","a","b","ab","c","ac","bc","abc"]
    **/
  static public <T> Iterable<Iterable<T>> subsequences(Iterable<T> xs){
    return cons(new Nil<T>(),nonEmptySubsequences(xs));
  }

  static private <T> Iterable<Iterable<T>> nonEmptySubsequences(Iterable<T> ys){
    if (ys.isEmpty()) return new Nil<Iterable<T>>();    
    T x = head(ys);
    Iterable<T> xs = tail(ys);
    return cons(cons(x,new Nil<T>())
               ,foldr( (Iterable<T> zs,Iterable<Iterable<T>>r) -> cons(zs ,cons( cons(x , zs) , r))
                     ,new Nil<Iterable<T>>()
                     ,nonEmptySubsequences(xs)));
  }

    /**
-- | The 'permutations' function returns the list of all permutations of the argument.
--
-- > permutations "abc" == ["abc","bac","cba","bca","cab","acb"]
permutations            :: [a] -> [[a]]
permutations xs0        =  xs0 : perms xs0 []
  where
    perms []     _  = []
    perms (t:ts) is = foldr interleave (perms ts (t:is)) (permutations is)
      where interleave    xs     r = let (_,zs) = interleave' id xs r in zs
            interleave' _ []     r = (ts, r)
            interleave' f (y:ys) r = let (us,zs) = interleave' (f . (y:)) ys r
                                     in  (y:us, f (t:y:us) : zs)
    **/
  static public <A> Iterable<Iterable<A>> permutations(Iterable<A> xs0){
    Function2<Iterable<A>,Iterable<A>,Iterable<Iterable<A>>> perms =
    (tso,is) -> {
      if (tso.isEmpty()) return new Nil<Iterable<A>>();
      A t = head(tso);
      Iterable<A> ts = tail(tso);

      return foldr((xs,r)->interleave_(t,ts,(u)->u,xs,r).e2,perms.eval(ts,cons(t,is)),permutations(is));
    };
    return cons(xs0,perms.eval(xs0,new Nil<A>()));
  }
  static private <A> Tuple2<Iterable<A>,Iterable<Iterable<A>>> interleave_ 
     (A t,Iterable<A> ts,Function1<Iterable<A> ,Iterable<A>>f ,Iterable<A> xs,Iterable<Iterable<A>> r){
    if (xs.isEmpty()) return new Tuple2<>(ts,r);
    A y = head(xs);
    Iterable<A> ys = tail(xs);
    Tuple2<Iterable<A>,Iterable<Iterable<A>>> uszs=interleave_(t,ts,(x)->f.eval(cons(y,x)),ys,r);
    return new Tuple2<Iterable<A>,Iterable<Iterable<A>>>(cons(y,uszs.e1),cons(f.eval(cons(t,cons(y,uszs.e1))),uszs.e2));
  }

  static public <A> Iterable<Iterable<A>> permutations2(Iterable<A> xs0){
    Function2<Iterable<A>,Iterable<A>,Iterable<Iterable<A>>> perms =
    (tso,is) -> {
      if (tso.isEmpty()) return new Nil<Iterable<A>>();
      Function5<A,Iterable<A>,Function1<Iterable<A> ,Iterable<A>>,Iterable<A>,Iterable<Iterable<A>>,Tuple2<Iterable<A>,Iterable<Iterable<A>>>> interleave_ =
      (t,ts,f,xs,r) -> {
        if (xs.isEmpty()) return new Tuple2<>(ts,r);
        A y = head(xs);
        Iterable<A> ys = tail(xs);
        Tuple2<Iterable<A>,Iterable<Iterable<A>>> uszs=interleave_.eval(t,ts,(x)->f.eval(cons(y,x)),ys,r);
        return new Tuple2<Iterable<A>,Iterable<Iterable<A>>>(cons(y,uszs.e1),cons(f.eval(cons(t,cons(y,uszs.e1))),uszs.e2));
      };
      A t = head(tso);
      Iterable<A> ts = tail(tso);

      return foldr((xs,r)->interleave_.eval(t,ts,(u)->u,xs,r).e2,perms.eval(ts,cons(t,is)),permutations2(is));
    };
    return cons(xs0,perms.eval(xs0,new Nil<A>()));
  }

  static public <A> Iterable<Iterable<A>> permutations3(Iterable<A> xs0){
    Function2<Iterable<A>,Iterable<A>,Iterable<Iterable<A>>> perms =
    (tso,is) -> {
      if (tso.isEmpty()) return new Nil<Iterable<A>>();
      A t = head(tso);
      Iterable<A> ts = tail(tso);

      Function3<Function1<Iterable<A> ,Iterable<A>>,Iterable<A>,Iterable<Iterable<A>>,Tuple2<Iterable<A>,Iterable<Iterable<A>>>> interleave_ =
      (f,xs,r) -> {
        if (xs.isEmpty()) return new Tuple2<Iterable<A>,Iterable<Iterable<A>>>(ts,r);
        A y = head(xs);
        Iterable<A> ys = tail(xs);
        Tuple2<Iterable<A>,Iterable<Iterable<A>>> uszs=interleave_.eval((x)->f.eval(cons(y,x)),ys,r);
        return new Tuple2<Iterable<A>,Iterable<Iterable<A>>>(cons(y,uszs.e1),cons(f.eval(cons(t,cons(y,uszs.e1))),uszs.e2));
      };

      return foldr((xs,r)->interleave_.eval((u)->u,xs,r).e2,perms.eval(ts,cons(t,is)),permutations(is));
    };
    return cons(xs0,perms.eval(xs0,new Nil<A>()));
  }


    /**
foldl, applied to a binary operator, a starting value (typically the left-identity of the operator), and a list, reduces the list using the binary operator, from left to right:

 foldl f z [x1, x2, ..., xn] == (...((z `f` x1) `f` x2) `f`...) `f` xn

The list must be finite. 
     **/
  static public <T1,T2> T2 foldl(Function2<T2,T1,T2> f,final T2 start,final Iterable<T1> xs){
    Iterator<T1> it = xs.iterator();
    T2 result = start;
    while (it.hasNext()) result=f.eval(result,it.next());
    return result;
  }
 
    /**
foldl1 is a variant of foldl that has no starting value argument, and thus must be applied to non-empty lists. 
     **/
  static public <T> T foldl(BinaryOperator<T> f,final Iterable<T> xs){
    Iterator<T> it = xs.iterator();
    T result = it.next();
    while (it.hasNext()) result=f.eval(result,it.next());
    return result;
  }


    /**
foldr, applied to a binary operator, a starting value (typically the right-identity of the operator), and a list, reduces the list using the binary operator, from right to left:

 foldr f z [x1, x2, ..., xn] == x1 `f` (x2 `f` ... (xn `f` z)...)


foldr            :: (a -> b -> b) -> b -> [a] -> b
foldr f z []     =  z
foldr f z (x:xs) =  f x (foldr f z xs)
    **/
  static public <T1,T2> T2 foldr(Function2<T1,T2,T2> f,final T2 start,final Iterable<T1> xs){
      if (xs.isEmpty()) return start;
      return f.eval(head(xs),foldr(f,start,tail(xs)));
  }
      /**

foldr1 is a variant of foldr that has no starting value argument, and thus must be applied to non-empty lists. 

foldr1           :: (a -> a -> a) -> [a] -> a
foldr1 f [x]     =  x
foldr1 f (x:xs)  =  f x (foldr1 f xs)
foldr1 _ []      =  error "Prelude.foldr1: empty list" **/
  static public <T> T foldr(BinaryOperator<T> f,final Iterable<T> xs){
    if (xs.isEmpty()) throw new HeadForEmptyIterableException("Prelude.foldr: empty list");
    T y = head(xs);
    Iterable<T> ys = tail(xs);
    if(ys.isEmpty()) return y;
    return f.eval(y,foldr(f,ys));
  }

    /**
Concatenate a list of lists. 
     **/
  static public <T> Iterable<T> concat(Iterable<Iterable<T>> xss){
    return foldr((Iterable<T> xs,Iterable<T> ys)->append(xs,ys),new Nil<T>(),xss);
  }

    /**
Map a function over a list and concatenate the results. 
concatMap :: (a -> [b]) -> [a] -> [b]
     **/
  static public <A,B> Iterable<B> concatMap(Function1<A,Iterable<B>>f,Iterable<A> xs){
    return concat(map(f,xs));
  }

    /**
       and returns the conjunction of a Boolean list. For the result to be True, the list must be finite; False, however, results from a False value at a finite index of a finite or infinite list. 
     **/
  static public <T> boolean and(Iterable<Boolean> xs){
    for (boolean b:xs) if(!b) return false;
    return true;
  }

    /**
       or returns the disjunction of a Boolean list. For the result to be False, the list must be finite; True, however, results from a True value at a finite index of a finite or infinite list. 
     **/
  static public <T> boolean or(Iterable<Boolean> xs){
    for (boolean b:xs) if(b) return true;
    return false;
  }

    /**
Applied to a predicate and a list, any determines if any element of the list satisfies the predicate. For the result to be False, the list must be finite; True, however, results from a True value for the predicate applied to an element at a finite index of a finite or infinite list. 
     **/
  static public <T> boolean any(Predicate<T> pred,final Iterable<T> xs){
    return !xs.filter(pred).isEmpty();
  }

    /**
Applied to a predicate and a list, all determines if all elements of the list satisfy the predicate. For the result to be True, the list must be finite; False, however, results from a False value for the predicate applied to an element at a finite index of a finite or infinite list. 
     **/
  static public <T> boolean all(Predicate<T> pred,final Iterable<T> xs){
    return !any((x)->!pred.test(x),xs);
  }


    /**The sum function computes the sum of a finite list of numbers. **/

  static public <T> int sum(Iterable<Integer> xs){
    int result=0;
    for (int i:xs) result=result+i;
    return result;
  }
    /** The product function computes the product of a finite list of numbers.  **/
  static public <T> int product(Iterable<Integer> xs){
    int result=1;
    for (int i:xs) result=result*i;
    return result;
  }


    /*
maximum returns the maximum value from a list, which must be non-empty, finite, and of an ordered type. It is a special case of maximumBy, which allows the programmer to supply their own comparison function.



maximum                 :: (Ord a) => [a] -> a
maximum []              =  errorEmptyList "maximum"
maximum xs              =  foldl1 max xs
    */
  static public <T extends Comparable<T>> T maximum(final Iterable<T> xs){
    if (xs.isEmpty()) throw new RuntimeException("no maximum element in empty list");
    return foldl(Iterables::max,xs);
  }

  static public <T> T maximum(Comparator<T>cmp,final Iterable<T> xs){
    if (xs.isEmpty()) throw new RuntimeException("no maximum element in empty list");
    return foldl((x,y)->cmp.compare(x,y)>0?x:y,xs);
  }

  static public <T extends Comparable<T>> T max(T x,T y){
    return x.compareTo(y)>0?x:y;
  }

  static public <T extends Comparable<T>> T minimum(final Iterable<T> xs){
    if (xs.isEmpty()) throw new RuntimeException("no minimum element in empty list");
    return foldl(Iterables::min,xs);
  }

  static public <T extends Comparable<T>> T min(T x,T y){
    return x.compareTo(y)<0?x:y;
  }

  static public <T> T minimum(Comparator<T>cmp,final Iterable<T> xs){
    if (xs.isEmpty()) throw new RuntimeException("no maximum element in empty list");
    return foldl((x,y)->cmp.compare(x,y)<0?x:y,xs);
  }


    /**scanl is similar to foldl, but returns a list of successive reduced values from the left:

 scanl f z [x1, x2, ...] == [z, z `f` x1, (z `f` x1) `f` x2, ...]

Note that

 last (scanl f z xs) == foldl f z xs.

scanl :: (a -> b -> a) -> a -> [b] -> [a]
    **/
  public static <A,B> Iterable<A> scanl(Function2<A,B,A>f,A q,Iterable<B> xs){
    return cons(q,xs.isEmpty()?new Nil<A>():scanl(f,f.eval(q,head(xs)),tail(xs)));
  }

    /**
       scanl1 is a variant of scanl that has no starting value argument: 
    **/
  public static <A> Iterable<A> scanl(Function2<A,A,A>f,Iterable<A> xs){
    if (xs.isEmpty()) return xs;
    return scanl(f, head(xs),tail(xs));
  }


  public static <A,B> Iterable<B> scanr(Function2<A,B,B>f,B q,Iterable<A> xs){
    if (xs.isEmpty()) return cons(q,new Nil<B>());
    A y = head(xs);
    Iterable<A> ys = tail(xs);
    Iterable<B> qs = scanr(f,q,ys);
    return  cons(f.eval(y,head(qs)),qs);
  }

	/**
scanr1          :: (a -> a -> a) -> [a] -> [a]
scanr1 f []     =  []
scanr1 f [x]    =  [x]
scanr1 f (x:xs) =  f x q : qs
                   where qs@(q:_) = scanr1 f xs 
	**/
  public static <A> Iterable<A> scanr(Function2<A,A,A>f,Iterable<A> xs){
    if (xs.isEmpty()) return new Nil<A>();
    Iterable<A> ys = tail(xs);
    if (ys.isEmpty()) return xs;
    A y = head(xs);
    Iterable<A> qs = scanr(f,ys);
    return cons(f.eval(y,head(qs)),qs);
  }

    /**
-- | The 'mapAccumL' function behaves like a combination of 'map' and
-- 'foldl'; it applies a function to each element of a list, passing
-- an accumulating parameter from left to right, and returning a final
-- value of this accumulator together with the new list.
mapAccumL :: (acc -> x -> (acc, y)) -- Function of elt of input list
                                    -- and accumulator, returning new
                                    -- accumulator and elt of result list
          -> acc            -- Initial accumulator 
          -> [x]            -- Input list
          -> (acc, [y])     -- Final accumulator and result list
mapAccumL _ s []        =  (s, [])
mapAccumL f s (x:xs)    =  (s'',y:ys)
                           where (s', y ) = f s x
                                 (s'',ys) = mapAccumL f s' xs
    **/
  static public <ACC,X,Y>  Tuple2<ACC,Iterable<Y>> mapAccumL(Function2<ACC,X,Tuple2<ACC,Y>> f,ACC acc,Iterable<X> xsA){
    if(xsA.isEmpty()) return new Tuple2<ACC,Iterable<Y>>(acc,new Nil<Y>());
    X x = head(xsA);
    Iterable<X> xs = tail(xsA);
    Tuple2<ACC,Y> sy = f.eval(acc,x);
    Tuple2<ACC,Iterable<Y>> sys = mapAccumL(f,sy.e1,xs);
    return new Tuple2<ACC,Iterable<Y>>(sys.e1,cons(sy.e2,sys.e2));
  }

    /**
mapAccumR _ s []        =  (s, [])
mapAccumR f s (x:xs)    =  (s'', y:ys)
                           where (s'',y ) = f s' x
                                 (s', ys) = mapAccumR f s xs
     **/
  static public <ACC,X,Y>  Tuple2<ACC,Iterable<Y>> mapAccumR(Function2<ACC,X,Tuple2<ACC,Y>> f,ACC acc,Iterable<X> xsA){
    if(xsA.isEmpty()) return new Tuple2<ACC,Iterable<Y>>(acc,new Nil<Y>());
    X x = head(xsA);
    Iterable<X> xs = tail(xsA);
    Tuple2<ACC,Iterable<Y>> sys = mapAccumR(f,acc,xs);
    Tuple2<ACC,Y> sy = f.eval(sys.e1,x);
    return new Tuple2<ACC,Iterable<Y>>(sy.e1,cons(sy.e2,sys.e2));
  }

    /**
-- | The 'unfoldr' function is a \`dual\' to 'foldr': while 'foldr'
-- reduces a list to a summary value, 'unfoldr' builds a list from
-- a seed value.  The function takes the element and returns 'Nothing'
-- if it is done producing the list or returns 'Just' @(a,b)@, in which
-- case, @a@ is a prepended to the list and @b@ is used as the next
-- element in a recursive call.  For example,
--
-- > iterate f == unfoldr (\x -> Just (x, f x))
--
-- In some cases, 'unfoldr' can undo a 'foldr' operation:
--
-- > unfoldr f' (foldr f z xs) == xs
--
-- if the following holds:
--
-- > f' (f x y) = Just (x,y)
-- > f' z       = Nothing
--
-- A simple use of unfoldr:
--
-- > unfoldr (\b -> if b == 0 then Nothing else Just (b, b-1)) 10
-- >  [10,9,8,7,6,5,4,3,2,1]
--
unfoldr      :: (b -> Maybe (a, b)) -> b -> [a]
unfoldr f b  =
  case f b of
   Just (a,new_b) -> a : unfoldr f new_b
   Nothing        -> []
    **/
  static public <A,B> Iterable<A> unfoldr(Function1<B,Tuple2<A,B>> f,B b){
    Tuple2<A,B> fb = f.eval(b);
    if (null==fb) return new Nil<A>();
    return cons(fb.e1,unfoldr(f,fb.e2));
  }


    /**
iterate f x returns an infinite list of repeated applications of f to x:

 iterate f x == [x, f x, f (f x), ...]
     **/
  static public <T> Iterable<T> iterate(final Function1<T,T> f,final T x){
    Iterator<T> result = new Iterator<T>(){
      T next = x;
      public boolean hasNext(){return true;}
      public T next(){
        T result = next; 
        next = f.eval(next);
        return result;
      }
    };
    return make(result);       
  }

    /**
repeat x is an infinite list, with x the value of every element. 
     **/
  static public <T> Iterable<T> repeat(final T x){
    Iterator<T> result = new Iterator<T>(){
      public boolean hasNext(){return true;}
      public T next(){return x;}
    };
    return make(result);
  }

    /**
replicate n x is a list of length n with x the value of every element.
     **/
  static public <T> Iterable<T> replicate(int i,final T x){
    return take(i,repeat(x));
  }
    /**

cycle ties a finite list into a circular one, or equivalently, the infinite repetition of the original list. It is the identity on infinite lists. 
     **/
  static public <T> Iterable<T> cycle(final Iterable<T> xs){
    Iterator<T> result = new Iterator<T>(){
    Iterator<T> it = xs.iterator();
    final boolean hasNext = it.hasNext();
    public boolean hasNext(){return hasNext;}
    public T next(){
      if (!it.hasNext()) it = xs.iterator();
        return it.next();
      }
    };
    return make(result);
  }


  static public <T> Iterable<T> take(final int i,final Iterable<T> xs){
    Iterator<T> result = new Iterator<T>(){
      Iterator<T> it = xs.iterator();
      int countDown = i;
      public boolean hasNext(){return it.hasNext()&&countDown>0;}
      public T next(){countDown--;return it.next();}
    };
    return make(result);
  }

  static public <T> Iterable<T> drop(int i,final Iterable<T> xs){
    Iterator<T> it = xs.iterator();
    for (;i>0;i--){
      if (it.hasNext()) it.next();
    }
    return make(it);
  }


  static public <A> UTuple2<Iterable<A>> splitAt(int i,final Iterable<A> xs){
    return new UTuple2<Iterable<A>>(take(i,xs),drop(i,xs));
  }

  static public <T> Iterable<T> takeWhile(final Predicate<T> pred,final Iterable<T> xs){
    if (xs.isEmpty()) return new Nil<T>();
    T x = head(xs);
    return pred.test(x)?cons(x,takeWhile(pred,tail(xs))):new Nil<T>();
  }

  static public <T> Iterable<T> dropWhile(final Predicate<T> pred,final Iterable<T> xs){
    Iterator<T> it = xs.iterator();
    T theNext=null;
    boolean hasNext=false;
    while (it.hasNext()){
      theNext=it.next();
      if (!pred.test(theNext)){
	  hasNext=true;
	  break;
      }
    }
    return (hasNext)?cons(theNext,make(it)):make(it);
  }

  static public <T> Iterable<T> dropWhileEnd(final Predicate<T> pred,final Iterable<T> xs){
    if (!all(pred,xs)) return cons(head(xs),dropWhileEnd(pred,tail(xs)));
    return new Nil<T>();
  }

  static public <A> UTuple2<Iterable<A>> span(Predicate<A> pred,final Iterable<A> xs){
    return new UTuple2<Iterable<A>>(takeWhile(pred,xs),dropWhile(pred,xs));
  }

  static public <A> UTuple2<Iterable<A>> breakIt(Predicate<A> pred,final Iterable<A> xs){
      return span((x)->!pred.test(x),xs);
  }

    /**
stripPrefix :: Eq a => [a] -> [a] -> Maybe [a]

The stripPrefix function drops the given prefix from a list. It returns Nothing if the list did not start with the prefix given, or Just the list after the prefix, if it does. 
    **/
  static public <T> Iterable<T> stripPrefix(final Iterable<T> xs,final Iterable<T> ys){ 
    if (xs.isEmpty()) return ys;
    if (ys.isEmpty()||!head(xs).equals(head(ys))) return null;
    return stripPrefix(tail(xs),tail(ys));
  }   


    /**
group :: Eq a => [a] -> [[a]]

The group function takes a list and returns a list of lists such that the concatenation of the result is equal to the argument. Moreover, each sublist in the result contains only equal elements. For example,

 group "Mississippi" = ["M","i","ss","i","ss","i","pp","i"]

It is a special case of groupBy, which allows the programmer to supply their own equality test. 
    **/
  static public <A> Iterable<Iterable<A>>group(Iterable<A> xs){
      return group((x,y)->x.equals(y),xs);
  }

    /**
-- | The 'groupBy' function is the non-overloaded version of 'group'.
groupBy                 :: (a -> a -> Bool) -> [a] -> [[a]]
groupBy _  []           =  []
groupBy eq (x:xs)       =  (x:ys) : groupBy eq zs
                           where (ys,zs) = span (eq x) xs

     **/
  static public <A> Iterable<Iterable<A>>group(Function2<A,A,Boolean> eq,Iterable<A> xs){
      if (xs.isEmpty()) return new Nil<Iterable<A>>();
      A x = head(xs);
      UTuple2<Iterable<A>> yszs = span((z)->eq.eval(z,x),tail(xs));
      return cons(cons(x,yszs.e1),group(eq,yszs.e2));
  }

    /**
The inits function returns all initial segments of the argument, shortest first. For example,

 inits "abc" == ["","a","ab","abc"]

     **/
  static public <A> Iterable<Iterable<A>> inits(Iterable<A> xs){
      return cons(new Nil<A>(),xs.isEmpty()?new Nil<Iterable<A>>():map((zs)->cons(head(xs),zs),inits(tail(xs))));
  }

    /**
The tails function returns all final segments of the argument, longest first. For example,

 tails "abc" == ["abc", "bc", "c",""]
    **/
  static public <A> Iterable<Iterable<A>> tails(Iterable<A> xs){
      return cons (xs,xs.isEmpty()?new Nil<Iterable<A>>():tails(tail(xs)));
  }

  static public <A> boolean isPrefixOf(Iterable<A> xs,Iterable<A> ys){
    if (xs.isEmpty()) return true;
    if (ys.isEmpty()) return false;
    return head(xs).equals(head(ys)) && isPrefixOf(tail(xs),tail(ys));
  }

  static public <A> boolean isSuffixOf(Iterable<A> xs,Iterable<A> ys){
    return isPrefixOf(reverse(xs),reverse(ys));
  }

  static public <A> boolean isInfixOf(Iterable<A> xs,Iterable<A> ys){
      return any((y)->isPrefixOf( xs,y),tails( ys));
  }

  static public <A> boolean elem(A a,Iterable<A> xs){
    if (xs.isEmpty()) return false;
    return head(xs).equals(a) || elem(a,tail(xs));
  }

  static public <A> boolean notElem(A a,Iterable<A> xs){  
    return all((y)->!y.equals(a),xs);
  }

  static public <A,B> B lookUp(A a,Iterable<Tuple2<A,B>> xs){
    if (xs.isEmpty()) return null;
    if (head(xs).e1.equals(a)) return head(xs).e2;
    return lookUp(a,tail(xs));
  }

  static public <A> A find(Predicate<A> p,Iterable<A> xs){  
    Iterable<A>  fs = filter(p,xs);
    return fs.isEmpty()?null:head(fs);
  }

  static public <A> Iterable<A> filter(Predicate<A> p,Iterable<A> xs){  
      return xs.filter(p);
  }

    /**
-- | The 'partition' function takes a predicate a list and returns
-- the pair of lists of elements which do and do not satisfy the
-- predicate, respectively; i.e.,
    **/
  static public <T> UTuple2<Iterable<T>> partition(Predicate<T> p,Iterable<T> xs){
    if (xs.isEmpty()) return new UTuple2<>(xs,xs);
    T x = head(xs);
    Iterable<T> txs = tail(xs);
    UTuple2<Iterable<T>> rek = partition(p,txs);
    return p.test(x)?new UTuple2<>(cons(x,rek.e1),rek.e2):new UTuple2<>(rek.e1,cons(x,rek.e2)); 
  }

  static public <A> A get(int i,Iterable<A> xs){  
      if (xs.isEmpty()) throw new HeadForEmptyIterableException("Prelude.get: empty list");
      if (i<0) throw new IndexOutOfBoundsException("Prelude.get: negativ index: "+i);
      if (i==0)return head(xs);
      return get(i-1,tail(xs));
  }

  static public <A> Integer elemIndex(A a,Iterable<A> xs){  
    return findIndex((x)-> x.equals(a),xs);
  }

  static public <A> Iterable<Integer> elemIndices(A a,Iterable<A> xs){  
      return findIndices((x)->a.equals(x),xs);
  }
    /**
-- | The 'findIndex' function takes a predicate and a list and returns
-- the index of the first element in the list satisfying the predicate,
-- or 'Nothing' if there is no such element.
findIndex       :: (a -> Bool) -> [a] -> Maybe Int
findIndex p     = listToMaybe . findIndices p
    **/
  static public <A> Integer findIndex(Predicate<A> p,Iterable<A> xs){  
      Iterable<Integer> is = findIndices(p,xs);
    return is.isEmpty()?null:head(is);
  }
    /**
-- | The 'findIndices' function extends 'findIndex', by returning the
-- indices of all elements satisfying the predicate, in ascending order.
findIndices      :: (a -> Bool) -> [a] -> [Int]
    **/
  static public <A> Iterable<Integer> findIndices(Predicate<A> pred,Iterable<A> xs){  
    return map((p)->p.e2,filter((p)->pred.test(p.e1),zip(xs,iterate((x)->x+1,0))));
  }

  static public <T1,T2> Iterable<Tuple2<T1,T2>> zip(final Iterable<T1> xs,final Iterable<T2> ys){
    return zipWith((x,y)->new Tuple2<T1,T2>(x,y),xs,ys);
  }

  static public <T1,T2,T3> Iterable<T3> zipWith(final Function2<T1,T2,T3>f,final Iterable<T1> xs,final Iterable<T2> ys){
    Iterator<T3> result = new Iterator<T3>(){
      Iterator<T1> it1 = xs.iterator();
      Iterator<T2> it2 = ys.iterator();
      public boolean hasNext(){return it1.hasNext()&&it2.hasNext();}
      public T3 next(){return f.eval(it1.next(),it2.next());}
    };
    return make(result);
  }

    /**
unzip transforms a list of pairs into a list of first components and a list of second components. 
     **/
  static public <A,B> Tuple2<Iterable<A>,Iterable<B>> unzip(final Iterable<Tuple2<A,B>> xys){
    return 
     foldr((ab,asbs)-> new Tuple2<Iterable<A>,Iterable<B>>(cons(ab.e1,asbs.e1),cons(ab.e2,asbs.e2))
          ,new Tuple2<Iterable<A>,Iterable<B>>(new Nil<A>(),new Nil<B>())
          ,xys);
  }

    static public Iterable<Iterable<Character>> lines(String str){
	return lines(mkStringIter(str));
    }

    static public Iterable<Iterable<Character>> lines(Iterable<Character> s){
      UTuple2<Iterable<Character>> ls =  breakIt((x)->x.equals('\n'),s);
      return cons(ls.e1,ls.e2.isEmpty()?new Nil<Iterable<Character>>():lines(tail(ls.e2)));
    }

    static public Iterable<Iterable<Character>> words(String str){
	return words(mkStringIter(str));
    }

    /**
words s                 =  case dropWhile {-partain:Char.-}isSpace s of
                                "" -> []
                                s' -> w : words s''
                                      where (w, s'') =
                                             break {-partain:Char.-}isSpace s'

     **/
  static public Iterable<Iterable<Character>> words(Iterable<Character> s){
    Iterable<Character> ls =  dropWhile(Character::isWhitespace,s);
    if (ls.isEmpty()) return new Nil<>();
    UTuple2<Iterable<Character>> ws =  breakIt(Character::isWhitespace,ls);
    return cons(ws.e1,words(ws.e2));
  }
  
  static public Iterable<Character> unlines( Iterable<Iterable<Character>> lns){
      return concatMap((ln)->append(ln,mkStringIter("\n")),lns);
  }

  static public Iterable<Character> unwords( Iterable<Iterable<Character>> ws){
      if (ws.isEmpty()) return new Nil<>();
      return foldr((w,s) -> append(w, cons(' ',s)), ws);
  }

  static public <A> Iterable<A> nub(Iterable<A> xs){
      return nub((x,y)->x.equals(y),xs); 
  }

  static public <A> Iterable<A> nub(Function2<A,A,Boolean> eq,Iterable<A> xs){
    if (xs.isEmpty()) return xs;
    A x = head(xs);
    Iterable<A> ys = tail(xs);
    return cons(x , nub(eq,filter((y) -> !eq.eval( x,y), xs)));
  }

    /**
-- | 'delete' @x@ removes the first occurrence of @x@ from its list argument.
-- For example,
--
-- > delete 'a' "banana" == "bnana"
--
-- It is a special case of 'deleteBy', which allows the programmer to
-- supply their own equality test.

delete                  :: (Eq a) => a -> [a] -> [a]
delete                  =  deleteBy (==)
    **/
  static public <A> Iterable<A> delete(A a,Iterable<A> xs){
    return delete((x,y)->x.equals(y),a,xs);
  }
    /**
-- | The 'deleteBy' function behaves like 'delete', but takes a
-- user-supplied equality predicate.
deleteBy                :: (a -> a -> Bool) -> a -> [a] -> [a]
deleteBy _  _ []        = []
deleteBy eq x (y:ys)    = if x `eq` y then ys else y : deleteBy eq x ys
    **/

  static public <A> Iterable<A> delete(Function2<A,A,Boolean> eq, A a,Iterable<A> xs){
    if (xs.isEmpty()) return new Nil<A>();
    A y=head(xs);
    Iterable<A> ys = tail(xs);
    if (eq.eval(a,y)) return ys;
    return cons(y,delete(eq,a,ys));
  }

    /**
(\\) :: Eq a => [a] -> [a] -> [a]

The \\ function is list difference ((non-associative). In the result of xs \\ ys, the first occurrence of each element of ys in turn (if any) has been removed from xs. Thus

 (xs ++ ys) \\ xs == ys.

It is a special case of deleteFirstsBy, which allows the programmer to supply their own equality test. 
    **/

  static public <A> Iterable<A> diff(Iterable<A> xs,Iterable<A> ys){
    return foldl((x,y)->delete(y,x),xs,ys);
  }

    /**
The union function returns the list union of the two lists. For example,

 "dog" `union` "cow" == "dogcw"

 Duplicates, and elements of the first list, are removed from the the second list, but if the first list contains duplicates, so will the result. It is a special case of unionBy, which allows the programmer to supply their own equality test. **/
  static public <A> Iterable<A> union(Iterable<A> xs,Iterable<A> ys){
    return union((x,y)->x.equals(y),xs,ys);
  }

    /**
-- | The 'unionBy' function is the non-overloaded version of 'union'.
unionBy                 :: (a -> a -> Bool) -> [a] -> [a] -> [a]
unionBy eq xs ys        =  xs ++ foldl (flip (deleteBy eq)) (nubBy eq ys) xs
    **/
  static public <A> Iterable<A> union(Function2<A,A,Boolean> eq,Iterable<A> xs,Iterable<A> ys){
    return append(xs,foldl((x,y)->delete(eq,y,x),nub(eq,ys),xs));
  }

    /**
The intersect function takes the list intersection of two lists. For example,

 [1,2,3,4] `intersect` [2,4,6,8] == [2,4]

If the first list contains duplicates, so will the result.

 [1,2,2,3,4] `intersect` [6,4,4,2] == [2,2,4]

It is a special case of intersectBy, which allows the programmer to supply their own equality test. 
    **/
  static public <A> Iterable<A> intersect(Iterable<A> xs,Iterable<A> ys){
    return intersect((x,y)->x.equals(y),xs,ys);
  }

    /**
-- | The 'intersectBy' function is the non-overloaded version of 'intersect'.
intersectBy             :: (a -> a -> Bool) -> [a] -> [a] -> [a]
intersectBy _  [] _     =  []
intersectBy _  _  []    =  []
intersectBy eq xs ys    =  [x | x <- xs, any (eq x) ys]
    **/
  static public <A> Iterable<A> intersect(Function2<A,A,Boolean> eq,Iterable<A> xs,Iterable<A> ys){
    if (xs.isEmpty()) return xs;
    if (ys.isEmpty()) return ys;
    return filter((x)->any((y)->x.equals(y),ys),xs);
  }

    /**
deleteFirstsBy :: (a -> a -> Bool) -> [a] -> [a] -> [a]

The deleteFirstsBy function takes a predicate and two lists and returns the first list with the first occurrence of each element of the second list removed. 
    **/
  static public <A> Iterable<A> deleteFirstsBy(Function2<A,A,Boolean> eq,Iterable<A> xs,Iterable<A> ys){
    return foldl((x,y)->delete(eq,y,x),xs,ys);
  }

    /**-- | The non-overloaded version of 'insert'.
insertBy :: (a -> a -> Ordering) -> a -> [a] -> [a]
insertBy _   x [] = [x]
insertBy cmp x ys@(y:ys')
 = case cmp x y of
     GT -> y : insertBy cmp x ys'
     _  -> x : ys
    **/
  static public <T> Iterable<T> insert(Comparator<T> cmp, T x ,final Iterable<T> xs){
    if (xs.isEmpty()) return cons(x,xs);
    T y = head(xs);
    Iterable<T> ys = tail(xs);
    return cmp.compare(x,y)>0?cons(y,insert(cmp,x,ys)):cons(x,xs);
  }

  static public <T extends Comparable<T>> Iterable<T> insert(T x ,final Iterable<T> xs){
      return insert((x1,x2)->x1.compareTo(x2),x,xs);
  }


  static public <T extends Comparable<T>> Iterable<T> qsort(final Iterable<T> xs){
    return qsort((x,y)->x.compareTo(y),xs);
  }

  static public <T> Iterable<T> qsort(Comparator<T> cmp,final Iterable<T> xs){
    if (xs.isEmpty()) return xs;
    T head = xs.getFirst();
    UTuple2<Iterable<T>> parted = partition((x)->cmp.compare(x,head)<0,tail(xs));
    return append(qsort(cmp,parted.e1),cons(head,qsort(cmp,parted.e2)));
  }


  /******************************************************************************/
  public static Iterable<Character> fileIterable(String fileName){
    try{
      final Reader reader = new BufferedReader(new FileReader(fileName));
      Iterator<Character> result = new Iterator<Character>(){
        int next = reader.read();

        public boolean hasNext(){
         return next >= 0;
        }
        public Character next(){
          char c = (char)next;
          try{
            next = reader.read();
          }catch(IOException e){next = -1;}
          return c;
        }
      };
      return make(result);
    }catch (IOException e){
	throw new RuntimeException(e);
    }
  }
}